from typing import List

from typing_extensions import Annotated
from util import pydantic_ref_fix

from ninja import Body, Cookie, Form, Header, NinjaAPI, Path, Query, Schema
from ninja.testing import TestClient

api = NinjaAPI()


class FormData(Schema):
    x: int
    y: float


class Payload(Schema):
    t: int
    p: str


@api.post("/multi/{p}")
def multi_op(
    request,
    q: Annotated[str, Query(description="Query param")],
    p: Annotated[int, Path(description="Path param")],
    f: Annotated[FormData, Form(description="Form params")],
    c: Annotated[str, Cookie(description="Cookie params")],
):
    return {"q": q, "p": p, "f": f.dict(), "c": c}


@api.post("/query_list")
def query_list(
    request,
    q: Annotated[List[str], Query(description="User ID")],
):
    return {"q": q}


@api.post("/headers")
def headers(request, h: Annotated[str, Header()] = "some-default"):
    return {"h": h}


@api.post("/body")
def body_op(
    request, payload: Annotated[Payload, Body(examples=[{"t": 42, "p": "test"}])]
):
    return {"payload": payload}


client = TestClient(api)


def test_multi_op():
    response = client.post("/multi/42?q=1", data={"x": 1, "y": 2}, COOKIES={"c": "3"})
    assert response.status_code == 200, response.content
    assert response.json() == {
        "q": "1",
        "p": 42,
        "f": {"x": 1, "y": 2.0},
        "c": "3",
    }


def test_query_list():
    response = client.post("/query_list?q=1&q=2")
    assert response.status_code == 200, response.content
    assert response.json() == {"q": ["1", "2"]}


def test_body_op():
    response = client.post("/body", json={"t": 42, "p": "test"})
    assert response.status_code == 200, response.content
    assert response.json() == {"payload": {"p": "test", "t": 42}}


def test_headers():
    response = client.post("/headers", headers={"h": "test"})
    assert response.status_code == 200, response.content
    assert response.json() == {"h": "test"}


def test_openapi_schema():
    schema = api.get_openapi_schema()["paths"]
    print(schema)
    assert schema == {
        "/api/multi/{p}": {
            "post": {
                "operationId": "test_annotated_multi_op",
                "summary": "Multi Op",
                "parameters": [
                    {
                        "in": "query",
                        "name": "q",
                        "schema": {
                            "description": "Query param",
                            "title": "Q",
                            "type": "string",
                        },
                        "required": True,
                        "description": "Query param",
                    },
                    {
                        "in": "path",
                        "name": "p",
                        "schema": {
                            "description": "Path param",
                            "title": "P",
                            "type": "integer",
                        },
                        "required": True,
                        "description": "Path param",
                    },
                    {
                        "in": "cookie",
                        "name": "c",
                        "schema": {
                            "description": "Cookie params",
                            "title": "C",
                            "type": "string",
                        },
                        "required": True,
                        "description": "Cookie params",
                    },
                ],
                "responses": {200: {"description": "OK"}},
                "requestBody": {
                    "content": {
                        "application/x-www-form-urlencoded": {
                            "schema": {
                                "title": "FormParams",
                                "type": "object",
                                "properties": {
                                    "x": {"title": "X", "type": "integer"},
                                    "y": {"title": "Y", "type": "number"},
                                },
                                "required": ["x", "y"],
                            }
                        }
                    },
                    "required": True,
                },
            }
        },
        "/api/query_list": {
            "post": {
                "operationId": "test_annotated_query_list",
                "summary": "Query List",
                "parameters": [
                    {
                        "in": "query",
                        "name": "q",
                        "schema": {
                            "description": "User ID",
                            "items": {"type": "string"},
                            "title": "Q",
                            "type": "array",
                        },
                        "required": True,
                        "description": "User ID",
                    }
                ],
                "responses": {200: {"description": "OK"}},
            }
        },
        "/api/headers": {
            "post": {
                "operationId": "test_annotated_headers",
                "summary": "Headers",
                "parameters": [
                    {
                        "in": "header",
                        "name": "h",
                        "schema": {
                            "default": "some-default",
                            "title": "H",
                            "type": "string",
                        },
                        "required": False,
                    }
                ],
                "responses": {200: {"description": "OK"}},
            }
        },
        "/api/body": {
            "post": {
                "operationId": "test_annotated_body_op",
                "summary": "Body Op",
                "parameters": [],
                "responses": {200: {"description": "OK"}},
                "requestBody": {
                    "content": {
                        "application/json": {
                            "schema": pydantic_ref_fix({
                                "$ref": "#/components/schemas/Payload",
                                "examples": [{"p": "test", "t": 42}],
                            })
                        }
                    },
                    "required": True,
                },
            }
        },
    }



from ninja import Form, Schema


class Item(Schema):
    name: str
    description: str = None
    price: float
    quantity: int


@api.post("/items/{item_id}")
def update(request, item_id: int, q: str, item: Form[Item]):
    return {"item_id": item_id, "item": item.dict(), "q": q}



import pytest
from pydantic import BaseModel

from ninja import Form, Query, Router
from ninja.testing import TestClient


class SomeModel(BaseModel):
    i: int
    s: str
    f: float


class OtherModel(BaseModel):
    x: int
    y: int


class SelfReference(BaseModel):
    a: int = 123
    sibling: "SelfReference" = None


SelfReference.model_rebuild()


router = Router()


@router.post("/test1")
def view1(request, some: SomeModel):
    assert isinstance(some, SomeModel)
    return some


@router.post("/test2")
def view2(request, some: SomeModel, other: OtherModel):
    assert isinstance(some, SomeModel)
    assert isinstance(other, OtherModel)
    return {"some": some, "other": other}


@router.post("/test3")
def view3(request, some: "SomeModel"):
    assert isinstance(some, SomeModel)
    return some


@router.post("/test_form")
def view4(request, form: OtherModel = Form(...)):
    assert isinstance(form, OtherModel)
    return form


@router.post("/test_query")
def view4query(request, q: OtherModel = Query(...)):
    assert isinstance(q, OtherModel)
    return q


@router.post("/selfref")
def view5(request, obj: SelfReference):
    assert isinstance(obj, SelfReference)
    return obj


@router.post("/model-default")
def view6(request, obj: OtherModel = None):
    assert isinstance(obj, (OtherModel, None.__class__))
    return obj


@router.post("/model-default2")
def view7(request, obj: OtherModel = OtherModel(x=1, y=1)):
    assert isinstance(obj, OtherModel)
    return obj


client = TestClient(router)


@pytest.mark.parametrize(
    # fmt: off
    "path,kwargs,expected_response",
    [
        (
            "/test1",
            dict(json={"i": "1", "s": "foo", "f": "1.1"}),
            {"i": 1, "s": "foo", "f": 1.1},
        ),
        (
            "/test2",
            dict(
                json={
                    "some": {"i": "1", "s": "foo", "f": "1.1"},
                    "other": {"x": 1, "y": 2},
                }
            ),
            {"some": {"i": 1, "s": "foo", "f": 1.1}, "other": {"x": 1, "y": 2}},
        ),
        (
            "/test3",
            dict(json={"i": "1", "s": "foo", "f": "1.1"}),
            {"i": 1, "s": "foo", "f": 1.1},
        ),
        (
            "/test_form",
            dict(data={"x": "10000", "y": "20000"}),
            {"x": 10000, "y": 20000},
        ),
        (
            "/test_query?x=5&y=6",
            dict(json=None),
            {"x": 5, "y": 6},
        ),
        (
            "/selfref",
            dict(json={"a": "1"}),
            {"a": 1, "sibling": None},
        ),
        (
            "/selfref",
            dict(json={"a": "1", "sibling": {"a": 2}}),
            {"a": 1, "sibling": {"a": 2, "sibling": None}},
        ),
        (
            "model-default",
            dict(json=None),
            None,
        ),
        (
            "model-default2",
            dict(json=None),
            {"x": 1, "y": 1},
        ),
    ],
    # fmt: on
)
def test_models(path, kwargs, expected_response):
    response = client.post(path, **kwargs)
    assert response.status_code == 200, response.content
    assert response.json() == expected_response


def test_invalid_body():
    response = client.post("/test1", body="invalid")
    assert response.status_code == 400, response.content
    assert response.json() == {
        "detail": "Cannot parse request body",
    }



from django.core.files.uploadedfile import SimpleUploadedFile

from ninja import File, Form, NinjaAPI, UploadedFile
from ninja.testing import TestClient

api = NinjaAPI()


@api.post("/str_and_file")
def str_and_file(
    request,
    title: str = Form(...),
    description: str = Form(""),
    file: UploadedFile = File(...),
):
    return {"title": title, "data": file.read().decode()}


client = TestClient(api)


def test_files():
    file = SimpleUploadedFile("test.txt", b"data123")
    response = client.post(
        "/str_and_file",
        FILES={"file": file},
        POST={"title": "hello"},
    )
    assert response.status_code == 200
    assert response.json() == {"title": "hello", "data": "data123"}

    schema = api.get_openapi_schema()["paths"]["/api/str_and_file"]
    r_body = schema["post"]["requestBody"]

    assert r_body == {
        "content": {
            "multipart/form-data": {
                "schema": {
                    "title": "MultiPartBodyParams",
                    "type": "object",
                    "properties": {
                        "title": {"title": "Title", "type": "string"},
                        "description": {
                            "title": "Description",
                            "default": "",
                            "type": "string",
                        },
                        "file": {
                            "title": "File",
                            "type": "string",
                            "format": "binary",
                        },
                    },
                    "required": ["title", "file"],
                }
            }
        },
        "required": True,
    }



from functools import wraps

import pytest

from ninja import NinjaAPI, Router
from ninja.testing import TestClient


# Test decorators
def operation_decorator(func):
    """Decorator that adds a header after validation (operation level)"""

    @wraps(func)
    def wrapper(request, *args, **kwargs):
        result = func(request, *args, **kwargs)
        if isinstance(result, dict):
            result["operation_decorated"] = True
        return result

    return wrapper


def view_decorator(func):
    """Decorator that adds a header before validation (view level)"""

    @wraps(func)
    def wrapper(request, *args, **kwargs):
        response = func(request, *args, **kwargs)
        response["X-View-Decorator"] = "applied"
        return response

    return wrapper


def counter_decorator(func):
    """Decorator that counts calls"""

    @wraps(func)
    def wrapper(request, *args, **kwargs):
        wrapper.call_count = getattr(wrapper, "call_count", 0) + 1
        result = func(request, *args, **kwargs)
        if isinstance(result, dict):
            result["call_count"] = wrapper.call_count
        return result

    return wrapper


def test_router_add_decorator_operation_mode():
    """Test add_decorator on router with OPERATION mode"""
    api = NinjaAPI()
    router = Router()

    # Add decorator to router
    router.add_decorator(operation_decorator, mode="operation")

    @router.get("/test")
    def endpoint(request):
        return {"message": "test"}

    api.add_router("/", router)
    client = TestClient(api)

    response = client.get("/test")
    assert response.status_code == 200
    assert response.json() == {"message": "test", "operation_decorated": True}


def test_router_add_decorator_view_mode():
    """Test add_decorator on router with VIEW mode"""
    api = NinjaAPI()
    router = Router()

    # Add decorator to router
    router.add_decorator(view_decorator, mode="view")

    @router.get("/test")
    def endpoint(request):
        return {"message": "test"}

    api.add_router("/", router)
    client = TestClient(api)

    response = client.get("/test")
    assert response.status_code == 200
    assert response["X-View-Decorator"] == "applied"
    assert response.json() == {"message": "test"}


def test_api_add_decorator_operation_mode():
    """Test add_decorator on API with OPERATION mode"""
    api = NinjaAPI()

    # Add decorator to entire API
    api.add_decorator(operation_decorator, mode="operation")

    @api.get("/test1")
    def endpoint1(request):
        return {"message": "test1"}

    @api.get("/test2")
    def endpoint2(request):
        return {"message": "test2"}

    client = TestClient(api)

    # Both endpoints should be decorated
    response = client.get("/test1")
    assert response.status_code == 200
    assert response.json() == {"message": "test1", "operation_decorated": True}

    response = client.get("/test2")
    assert response.status_code == 200
    assert response.json() == {"message": "test2", "operation_decorated": True}


def test_api_add_decorator_view_mode():
    """Test add_decorator on API with VIEW mode"""
    api = NinjaAPI()

    # Add decorator to entire API
    api.add_decorator(view_decorator, mode="view")

    @api.get("/test")
    def endpoint(request):
        return {"message": "test"}

    client = TestClient(api)

    response = client.get("/test")
    assert response.status_code == 200
    assert response["X-View-Decorator"] == "applied"


def test_multiple_decorators():
    """Test multiple decorators on same router"""
    api = NinjaAPI()
    router = Router()

    # Add multiple decorators
    router.add_decorator(operation_decorator, mode="operation")
    router.add_decorator(counter_decorator, mode="operation")

    @router.get("/test")
    def endpoint(request):
        return {"message": "test"}

    api.add_router("/", router)
    client = TestClient(api)

    response = client.get("/test")
    assert response.status_code == 200
    result = response.json()
    assert result["message"] == "test"
    assert result["operation_decorated"] is True
    assert result["call_count"] == 1

    # Second call should increment counter
    response = client.get("/test")
    result = response.json()
    assert result["call_count"] == 2


def test_decorator_cascading():
    """Test that decorators cascade from API to router to child router"""
    api = NinjaAPI()
    parent_router = Router()
    child_router = Router()

    # Add decorator at API level
    api.add_decorator(
        lambda f: wraps(f)(lambda req, *a, **k: {**f(req, *a, **k), "api": True})
    )

    # Add decorator at parent router level
    parent_router.add_decorator(
        lambda f: wraps(f)(lambda req, *a, **k: {**f(req, *a, **k), "parent": True})
    )

    # Add decorator at child router level
    child_router.add_decorator(
        lambda f: wraps(f)(lambda req, *a, **k: {**f(req, *a, **k), "child": True})
    )

    @child_router.get("/test")
    def endpoint(request):
        return {"message": "test"}

    parent_router.add_router("/child", child_router)
    api.add_router("/parent", parent_router)

    client = TestClient(api)
    response = client.get("/parent/child/test")
    assert response.status_code == 200
    result = response.json()
    assert result == {
        "message": "test",
        "api": True,
        "parent": True,
        "child": True,
    }


def test_api_decorator_applies_to_new_routers():
    """Test that API-level decorators apply to routers added after decorator"""
    api = NinjaAPI()

    # Add decorator to API first
    api.add_decorator(operation_decorator, mode="operation")

    # Then add a router
    router = Router()

    @router.get("/test")
    def endpoint(request):
        return {"message": "test"}

    api.add_router("/", router)

    client = TestClient(api)
    response = client.get("/test")
    assert response.status_code == 200
    assert response.json() == {"message": "test", "operation_decorated": True}


def test_mix_view_and_operation_decorators():
    """Test mixing VIEW and OPERATION mode decorators"""
    api = NinjaAPI()
    router = Router()

    # Add both types of decorators
    router.add_decorator(view_decorator, mode="view")
    router.add_decorator(operation_decorator, mode="operation")

    @router.get("/test")
    def endpoint(request):
        return {"message": "test"}

    api.add_router("/", router)
    client = TestClient(api)

    response = client.get("/test")
    assert response.status_code == 200
    assert response["X-View-Decorator"] == "applied"
    assert response.json() == {"message": "test", "operation_decorated": True}


def test_decorator_with_path_params():
    """Test decorators work with path parameters"""
    api = NinjaAPI()
    router = Router()

    def param_decorator(func):
        @wraps(func)
        def wrapper(request, *args, **kwargs):
            result = func(request, *args, **kwargs)
            if isinstance(result, dict):
                result["decorated"] = True
            return result

        return wrapper

    router.add_decorator(param_decorator, mode="operation")

    @router.get("/test/{item_id}")
    def endpoint(request, item_id: int):
        return {"item_id": item_id}

    api.add_router("/", router)
    client = TestClient(api)

    response = client.get("/test/123")
    assert response.status_code == 200
    assert response.json() == {"item_id": 123, "decorated": True}


def test_invalid_decorator_mode():
    """Test that invalid decorator mode raises ValueError"""
    router = Router()

    with pytest.raises(ValueError, match="Invalid decorator mode"):
        router.add_decorator(operation_decorator, mode="invalid")  # type: ignore
